# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2020-12-25 18:48


from urllib.error import HTTPError

from hanlp.datasets.srl.ontonotes5 import ONTONOTES5_HOME, CONLL12_HOME
from hanlp.datasets.srl.ontonotes5._utils import make_gold_conll, make_ontonotes_language_jsonlines, \
    batch_make_ner_tsv_if_necessary, batch_make_pos_tsv_if_necessary, batch_make_con_txt_if_necessary, \
    batch_make_dep_conllx_if_necessary
from hanlp.utils.io_util import get_resource, path_from_url
from hanlp.utils.log_util import cprint

_ONTONOTES5_ENGLISH_HOME = ONTONOTES5_HOME + 'files/data/english/'
_ONTONOTES5_CONLL12_ENGLISH_HOME = CONLL12_HOME + 'english/'

ONTONOTES5_CONLL12_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.conll12.jsonlines'
'''Training set of English OntoNotes5 used in CoNLL12 (:cite:`pradhan-etal-2012-conll`).'''
ONTONOTES5_CONLL12_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.conll12.jsonlines'
'''Dev set of English OntoNotes5 used in CoNLL12 (:cite:`pradhan-etal-2012-conll`).'''
ONTONOTES5_CONLL12_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.conll12.jsonlines'
'''Test set of English OntoNotes5 used in CoNLL12 (:cite:`pradhan-etal-2012-conll`).'''

ONTONOTES5_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.v4.jsonlines'
ONTONOTES5_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.v4.jsonlines'
ONTONOTES5_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.v4.jsonlines'

ONTONOTES5_CONLL_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.v4_gold_conll'
ONTONOTES5_CONLL_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.v4_gold_conll'
ONTONOTES5_CONLL_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.v4_gold_conll'

ONTONOTES5_POS_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.v4.pos.tsv'
ONTONOTES5_POS_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.v4.pos.tsv'
ONTONOTES5_POS_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.v4.pos.tsv'

ONTONOTES5_CON_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.con.txt'
ONTONOTES5_CON_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.con.txt'
ONTONOTES5_CON_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.con.txt'

ONTONOTES5_DEP_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.dep.conllx'
ONTONOTES5_DEP_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.dep.conllx'
ONTONOTES5_DEP_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.dep.conllx'

# ONTONOTES5_CON_ENGLISH_NOEC_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.con.noempty.txt'
# ONTONOTES5_CON_ENGLISH_NOEC_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.con.noempty.txt'
# ONTONOTES5_CON_ENGLISH_NOEC_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.con.noempty.txt'

ONTONOTES5_CONLL12_NER_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.conll12.ner.tsv'
'''Training set of English OntoNotes5 used in CoNLL12 (:cite:`pradhan-etal-2012-conll`).'''
ONTONOTES5_CONLL12_NER_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.conll12.ner.tsv'
'''Dev set of English OntoNotes5 used in CoNLL12 (:cite:`pradhan-etal-2012-conll`).'''
ONTONOTES5_CONLL12_NER_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.conll12.ner.tsv'
'''Test set of English OntoNotes5 used in CoNLL12 (:cite:`pradhan-etal-2012-conll`).'''

ONTONOTES5_NER_ENGLISH_TRAIN = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'train.english.v4.ner.tsv'
ONTONOTES5_NER_ENGLISH_DEV = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'development.english.v4.ner.tsv'
ONTONOTES5_NER_ENGLISH_TEST = _ONTONOTES5_CONLL12_ENGLISH_HOME + 'test.english.v4.ner.tsv'

try:
    get_resource(ONTONOTES5_HOME, verbose=False)
except HTTPError:
    intended_file_path = path_from_url(ONTONOTES5_HOME)
    cprint('Ontonotes 5.0 is a [red][bold]copyright[/bold][/red] dataset owned by LDC which we cannot re-distribute. '
           f'Please apply for a licence from LDC (https://catalog.ldc.upenn.edu/LDC2016T13) '
           f'then download it to {intended_file_path}')
    exit(1)

try:
    get_resource(ONTONOTES5_CONLL12_ENGLISH_TRAIN, verbose=False)
except HTTPError:
    make_gold_conll(ONTONOTES5_HOME + '..', 'english')
    make_ontonotes_language_jsonlines(CONLL12_HOME + 'v4', language='english')

batch_make_ner_tsv_if_necessary(
    [ONTONOTES5_CONLL12_ENGLISH_TRAIN, ONTONOTES5_CONLL12_ENGLISH_DEV, ONTONOTES5_CONLL12_ENGLISH_TEST])

batch_make_ner_tsv_if_necessary(
    [ONTONOTES5_ENGLISH_TRAIN, ONTONOTES5_ENGLISH_DEV, ONTONOTES5_ENGLISH_TEST])

batch_make_pos_tsv_if_necessary(
    [ONTONOTES5_ENGLISH_TRAIN, ONTONOTES5_ENGLISH_DEV, ONTONOTES5_ENGLISH_TEST])

batch_make_con_txt_if_necessary(
    [ONTONOTES5_CONLL_ENGLISH_TRAIN, ONTONOTES5_CONLL_ENGLISH_DEV, ONTONOTES5_CONLL_ENGLISH_TEST])

batch_make_dep_conllx_if_necessary(
    [ONTONOTES5_CON_ENGLISH_TRAIN, ONTONOTES5_CON_ENGLISH_DEV, ONTONOTES5_CON_ENGLISH_TEST])

# batch_remove_empty_category_if_necessary(
#     [ONTONOTES5_CON_ENGLISH_TRAIN, ONTONOTES5_CON_ENGLISH_DEV, ONTONOTES5_CON_ENGLISH_TEST])



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-28 21:12
from hanlp_common.constant import HANLP_URL

SIGHAN2005_PKU_CONVSEG = HANLP_URL + 'tok/sighan2005-pku-convseg_20200110_153722.zip'
'Conv model (:cite:`wang-xu-2017-convolutional`) trained on sighan2005 pku dataset.'
SIGHAN2005_MSR_CONVSEG = HANLP_URL + 'tok/convseg-msr-nocrf-noembed_20200110_153524.zip'
'Conv model (:cite:`wang-xu-2017-convolutional`) trained on sighan2005 msr dataset.'
# SIGHAN2005_MSR_BERT_BASE = HANLP_URL + 'tok/cws_bert_base_msra_20191230_194627.zip'
CTB6_CONVSEG = HANLP_URL + 'tok/ctb6_convseg_nowe_nocrf_20200110_004046.zip'
'Conv model (:cite:`wang-xu-2017-convolutional`) trained on CTB6 dataset.'
# CTB6_BERT_BASE = HANLP_URL + 'tok/cws_bert_base_ctb6_20191230_185536.zip'
PKU_NAME_MERGED_SIX_MONTHS_CONVSEG = HANLP_URL + 'tok/pku98_6m_conv_ngram_20200110_134736.zip'
'Conv model (:cite:`wang-xu-2017-convolutional`) trained on pku98 six months dataset with name merged into one unit.'
LARGE_ALBERT_BASE = HANLP_URL + 'tok/large_cws_albert_base_20200828_011451.zip'
'ALBERT model (:cite:`Lan2020ALBERT:`) trained on the largest CWS dataset in the world.'
SIGHAN2005_PKU_BERT_BASE_ZH = HANLP_URL + 'tok/sighan2005_pku_bert_base_zh_20201231_141130.zip'
'BERT model (:cite:`devlin-etal-2019-bert`) trained on sighan2005 pku dataset.'

# Will be filled up during runtime
ALL = {}



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from hanlp.utils.tf_util import hanlp_register


@hanlp_register
class WeightNormalization(tf.keras.layers.Wrapper):
    """This wrapper reparameterizes a layer by decoupling the weight's
    magnitude and direction.
    
    This speeds up convergence by improving the
    conditioning of the optimization problem.
    Weight Normalization: A Simple Reparameterization to Accelerate
    Training of Deep Neural Networks: https://arxiv.org/abs/1602.07868
    Tim Salimans, Diederik P. Kingma (2016)
    WeightNormalization wrapper works for keras and tf layers.
    ```python
      net = WeightNormalization(
          tf.keras.layers.Conv2D(2, 2, activation='relu'),
          input_shape=(32, 32, 3),
          data_init=True)(x)
      net = WeightNormalization(
          tf.keras.layers.Conv2D(16, 5, activation='relu'),
          data_init=True)(net)
      net = WeightNormalization(
          tf.keras.layers.Dense(120, activation='relu'),
          data_init=True)(net)
      net = WeightNormalization(
          tf.keras.layers.Dense(n_classes),
          data_init=True)(net)
    ```

    Args:
      layer: a layer instance
      data_init: If

    Returns:

    Raises:
      ValueError: If not initialized with a
      ValueError: If
      NotImplementedError: If

    """

    def __init__(self, layer, data_init=True, **kwargs):
        super(WeightNormalization, self).__init__(layer, **kwargs)
        self.data_init = data_init
        self._track_trackable(layer, name='layer')
        self._init_critical_section = tf.CriticalSection(name='init_mutex')
        self.is_rnn = isinstance(self.layer, tf.keras.layers.RNN)

    def build(self, input_shape):
        """Build `Layer`

        Args:
          input_shape: 

        Returns:

        """
        input_shape = tf.TensorShape(input_shape)
        self.input_spec = tf.keras.layers.InputSpec(
            shape=[None] + input_shape[1:])

        if not self.layer.built:
            self.layer.build(input_shape)

        kernel_layer = self.layer.cell if self.is_rnn else self.layer

        if not hasattr(kernel_layer, 'kernel'):
            raise ValueError('`WeightNormalization` must wrap a layer that'
                             ' contains a `kernel` for weights')

        # The kernel's filter or unit dimension is -1
        self.layer_depth = int(kernel_layer.kernel.shape[-1])
        self.kernel_norm_axes = list(range(kernel_layer.kernel.shape.rank - 1))

        self.g = self.add_weight(
            name='g',
            shape=(self.layer_depth,),
            initializer='ones',
            dtype=kernel_layer.kernel.dtype,
            trainable=True)
        self.v = kernel_layer.kernel

        self._initialized = self.add_weight(
            name='initialized',
            shape=None,
            initializer='zeros',
            dtype=tf.dtypes.bool,
            trainable=False)

        if self.data_init:
            # Used for data initialization in self._data_dep_init.
            with tf.name_scope('data_dep_init'):
                layer_config = tf.keras.layers.serialize(self.layer)
                layer_config['config']['trainable'] = False
                self._naked_clone_layer = tf.keras.layers.deserialize(
                    layer_config)
                self._naked_clone_layer.build(input_shape)
                self._naked_clone_layer.set_weights(self.layer.get_weights())
                if self.is_rnn:
                    self._naked_clone_layer.cell.activation = None
                else:
                    self._naked_clone_layer.activation = None

        self.built = True

    def call(self, inputs):
        """Call `Layer`

        Args:
          inputs: 

        Returns:

        """

        def _do_nothing():
            return tf.identity(self.g)

        def _update_weights():
            # Ensure we read `self.g` after _update_weights.
            with tf.control_dependencies(self._initialize_weights(inputs)):
                return tf.identity(self.g)

        g = self._init_critical_section.execute(lambda: tf.cond(
            self._initialized, _do_nothing, _update_weights))

        with tf.name_scope('compute_weights'):
            # Replace kernel by normalized weight variable.
            self.layer.kernel = tf.nn.l2_normalize(
                self.v, axis=self.kernel_norm_axes) * g

            # Ensure we calculate result after updating kernel.
            update_kernel = tf.identity(self.layer.kernel)
            with tf.control_dependencies([update_kernel]):
                outputs = self.layer(inputs)
                return outputs

    def compute_output_shape(self, input_shape):
        return tf.TensorShape(
            self.layer.compute_output_shape(input_shape).as_list())

    def _initialize_weights(self, inputs):
        """Initialize weight g.
        
        The initial value of g could either from the initial value in v,
        or by the input value if self.data_init is True.

        Args:
          inputs: 

        Returns:

        """
        with tf.control_dependencies([
            tf.debugging.assert_equal(  # pylint: disable=bad-continuation
                self._initialized,
                False,
                message='The layer has been initialized.')
        ]):
            if self.data_init:
                assign_tensors = self._data_dep_init(inputs)
            else:
                assign_tensors = self._init_norm()
            assign_tensors.append(self._initialized.assign(True))
            return assign_tensors

    def _init_norm(self):
        """Set the weight g with the norm of the weight vector."""
        with tf.name_scope('init_norm'):
            v_flat = tf.reshape(self.v, [-1, self.layer_depth])
            v_norm = tf.linalg.norm(v_flat, axis=0)
            g_tensor = self.g.assign(tf.reshape(v_norm, (self.layer_depth,)))
            return [g_tensor]

    def _data_dep_init(self, inputs):
        """Data dependent initialization.

        Args:
          inputs: 

        Returns:

        """
        with tf.name_scope('data_dep_init'):
            # Generate data dependent init values
            x_init = self._naked_clone_layer(inputs)
            data_norm_axes = list(range(x_init.shape.rank - 1))
            m_init, v_init = tf.nn.moments(x_init, data_norm_axes)
            scale_init = 1. / tf.math.sqrt(v_init + 1e-10)

            # Assign data dependent init values
            g_tensor = self.g.assign(self.g * scale_init)
            if hasattr(self.layer, 'bias') and self.layer.bias is not None:
                bias_tensor = self.layer.bias.assign(-m_init * scale_init)
                return [g_tensor, bias_tensor]
            else:
                return [g_tensor]

    def get_config(self):
        config = {'data_init': self.data_init}
        base_config = super(WeightNormalization, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2020-03-08 22:35
import tempfile

from hanlp.components.parsers.conll import read_conll
from hanlp.utils.io_util import get_resource, get_exitcode_stdout_stderr

CONLLX_EVAL = get_resource(
    'https://github.com/elikip/bist-parser/archive/master.zip' + '#bmstparser/src/utils/eval.pl')


def evaluate(gold_file, pred_file):
    """Evaluate using official CoNLL-X evaluation script (Yuval Krymolowski)

    Args:
      gold_file(str): The gold conllx file
      pred_file(str): The pred conllx file

    Returns:

    
    """
    gold_file = get_resource(gold_file)
    fixed_pred_file = tempfile.NamedTemporaryFile().name
    copy_cols(gold_file, pred_file, fixed_pred_file, keep_comments=False)
    if gold_file.endswith('.conllu'):
        fixed_gold_file = tempfile.NamedTemporaryFile().name
        copy_cols(gold_file, gold_file, fixed_gold_file, keep_comments=False)
        gold_file = fixed_gold_file

    exitcode, out, err = get_exitcode_stdout_stderr(f'perl {CONLLX_EVAL} -q -b -g {gold_file} -s {fixed_pred_file}')
    if exitcode:
        raise RuntimeError(f'eval.pl exited with error code {exitcode} and error message {err} and output {out}.')
    lines = out.split('\n')[-4:]
    las = int(lines[0].split()[3]) / int(lines[0].split()[5])
    uas = int(lines[1].split()[3]) / int(lines[1].split()[5])
    return uas, las


def copy_cols(gold_file, pred_file, copied_pred_file, keep_comments=True):
    """Copy the first 6 columns from gold file to pred file

    Args:
      gold_file: 
      pred_file: 
      copied_pred_file: 
      keep_comments:  (Default value = True)

    Returns:

    
    """
    with open(copied_pred_file, 'w') as to_out, open(pred_file) as pred_file, open(gold_file) as gold_file:
        for idx, (p, g) in enumerate(zip(pred_file, gold_file)):
            while p.startswith('#'):
                p = next(pred_file)
            if not g.strip():
                if p.strip():
                    raise ValueError(
                        f'Prediction file {pred_file.name} does not end a sentence at line {idx + 1}\n{p.strip()}')
                to_out.write('\n')
                continue
            while g.startswith('#') or '-' in g.split('\t')[0]:
                if keep_comments or g.startswith('-'):
                    to_out.write(g)
                g = next(gold_file)
            to_out.write('\t'.join(str(x) for x in g.split('\t')[:6] + p.split('\t')[6:]))



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2020-07-05 19:34
from alnlp.modules.pytorch_seq2seq_wrapper import LstmSeq2SeqEncoder
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from hanlp.common.structure import ConfigTracker


class _LSTMSeq2Seq(nn.Module):
    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            num_layers: int = 1,
            bias: bool = True,
            dropout: float = 0.0,
            bidirectional: bool = False,
    ):
        """
        Under construction, not ready for production
        :param input_size:
        :param hidden_size:
        :param num_layers:
        :param bias:
        :param dropout:
        :param bidirectional:
        """
        self.rnn = nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=True,
            dropout=dropout,
            bidirectional=bidirectional,
        )

    def forward(self, embed, lens, max_len):
        x = pack_padded_sequence(embed, lens, True, False)
        x, _ = self.rnn(x)
        x, _ = pad_packed_sequence(x, True, total_length=max_len)
        return x


# We might update this to support yaml based configuration
class LSTMContextualEncoder(LstmSeq2SeqEncoder, ConfigTracker):

    def __init__(self, input_size: int, hidden_size: int, num_layers: int = 1, bias: bool = True, dropout: float = 0.0,
                 bidirectional: bool = False, stateful: bool = False):
        super().__init__(input_size, hidden_size, num_layers, bias, dropout, bidirectional, stateful)
        ConfigTracker.__init__(self, locals())



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2020-07-05 20:28
import logging
from typing import Union, List, Callable, Dict, Any, Tuple

import torch
from hanlp.layers.transformers.utils import get_optimizers
from alnlp.metrics.conll_coref_scores import ConllCorefScores
from alnlp.metrics.mention_recall import MentionRecall
from alnlp.models.coref import CoreferenceResolver
from alnlp.modules.initializers import InitializerApplicator
from alnlp.modules.util import lengths_to_mask
from alnlp.training.optimizers import make_parameter_groups
from torch.utils.data import DataLoader
from hanlp.common.dataset import PadSequenceDataLoader
from hanlp.common.torch_component import TorchComponent
from hanlp.common.transform import FieldLength
from hanlp.datasets.coref.conll12coref import CONLL12CorefDataset
from hanlp.layers.context_layer import LSTMContextualEncoder
from hanlp.layers.embeddings.embedding import Embedding
from hanlp.layers.feed_forward import FeedForward
from hanlp.utils.time_util import CountdownTimer
from hanlp.utils.torch_util import clip_grad_norm
from hanlp_common.util import merge_locals_kwargs


class CoreferenceResolverModel(CoreferenceResolver):
    # noinspection PyMethodOverriding
    def forward(self, batch: dict) -> Dict[str, torch.Tensor]:
        batch['mask'] = mask = lengths_to_mask(batch['text_length'])
        return super().forward(batch, batch['spans'], batch.get('span_labels'), mask=mask)


class EndToEndCoreferenceResolver(TorchComponent):
    def build_optimizer(self,
                        trn,
                        epochs,
                        lr,
                        adam_epsilon,
                        weight_decay,
                        warmup_steps,
                        transformer_lr,
                        **kwargs):
        # noinspection PyProtectedMember
        transformer = getattr(self.model._text_field_embedder, 'transformer', None)
        if transformer:
            model = self.model
            num_training_steps = len(trn) * epochs // self.config.get('gradient_accumulation', 1)

            optimizer_grouped_parameters = make_parameter_groups(list(self.model.named_parameters()),
                                                                 [([".*transformer.*"], {"lr": transformer_lr})])
            optimizer, linear_scheduler = get_optimizers(model,
                                                         num_training_steps,
                                                         learning_rate=lr,
                                                         adam_epsilon=adam_epsilon,
                                                         weight_decay=weight_decay,
                                                         warmup_steps=warmup_steps,
                                                         optimizer_grouped_parameters=optimizer_grouped_parameters
                                                         )
        else:
            optimizer = torch.optim.Adam(self.model.parameters(), self.config.lr)
            linear_scheduler = None
        reduce_lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer=optimizer,
            mode='max',
            factor=0.5,
            patience=2,
            verbose=True,
        )
        return optimizer, reduce_lr_scheduler, linear_scheduler

    def build_criterion(self, **kwargs):
        pass

    def build_metric(self, **kwargs) -> Tuple[MentionRecall, ConllCorefScores]:
        return self.model._mention_recall, self.model._conll_coref_scores

    def execute_training_loop(self,
                              trn: DataLoader,
                              dev: DataLoader,
                              epochs,
                              criterion,
                              optimizer,
                              metric,
                              save_dir,
                              logger: logging.Logger,
                              devices,
                              **kwargs):
        best_epoch, best_metric = 0, -1
        mention_recall, conll_coref_scores = self.build_metric()
        optimizer, reduce_lr_scheduler, linear_scheduler = optimizer
        timer = CountdownTimer(epochs)
        ratio_width = len(f'{len(trn)}/{len(trn)}')
        for epoch in range(1, epochs + 1):
            logger.info(f"[yellow]Epoch {epoch} / {epochs}:[/yellow]")
            self.fit_dataloader(trn, criterion, optimizer, metric, logger, linear_scheduler=linear_scheduler)
            if dev:
                self.evaluate_dataloader(dev, criterion, metric, logger, ratio_width=ratio_width)
            report = f'{timer.elapsed_human}/{timer.total_time_human}'
            dev_score = conll_coref_scores.get_metric()[-1]
            reduce_lr_scheduler.step(dev_score)
            if dev_score > best_metric:
                self.save_weights(save_dir)
                best_metric = dev_score
                report += ' [red]saved[/red]'
            timer.log(report, ratio_percentage=False, newline=True, ratio=False)

    def fit_dataloader(self,
                       trn: DataLoader,
                       criterion,
                       optimizer,
                       metric,
                       logger: logging.Logger,
                       linear_scheduler=None,
                       **kwargs):
        self.model.train()
        timer = CountdownTimer(len(trn))
        total_loss = 0
        self.reset_metrics()
        for batch in trn:
            optimizer.zero_grad()
            output_dict = self.feed_batch(batch)
            loss = output_dict['loss']
            loss.backward()
            if self.config.grad_norm:
                clip_grad_norm(self.model, self.config.grad_norm)
            optimizer.step()
            if linear_scheduler:
                linear_scheduler.step()
            total_loss += loss.item()
            timer.log(self.report_metrics(total_loss / (timer.current + 1)), ratio_percentage=None, logger=logger)
            del loss
        return total_loss / timer.total

    # noinspection PyMethodOverriding
    def evaluate_dataloader(self,
                            data: DataLoader,
                            criterion: Callable,
                            metric,
                            logger,
                            ratio_width=None,
                            output=False,
                            **kwargs):
        self.model.eval()
        self.reset_metrics()
        timer = CountdownTimer(len(data))
        total_loss = 0
        self.reset_metrics()
        for batch in data:
            output_dict = self.feed_batch(batch)
            loss = output_dict['loss']
            total_loss += loss.item()
            timer.log(self.report_metrics(total_loss / (timer.current + 1)), ratio_percentage=None, logger=logger,
                      ratio_width=ratio_width)
            del loss
        return total_loss / timer.total

    def build_model(self,
                    training=True,
                    **kwargs) -> torch.nn.Module:
        # noinspection PyTypeChecker
        model = CoreferenceResolverModel(
            self.config.embed.module(vocabs=self.vocabs, training=training),
            self.config.context_layer,
            self.config.mention_feedforward,
            self.config.antecedent_feedforward,
            self.config.feature_size,
            self.config.max_span_width,
            self.config.spans_per_word,
            self.config.max_antecedents,
            self.config.coarse_to_fine,
            self.config.inference_order,
            self.config.lexical_dropout,
            InitializerApplicator([
                [".*linear_layers.*weight", {"type": "xavier_normal"}],
                [".*scorer._module.weight", {"type": "xavier_normal"}],
                ["_distance_embedding.weight", {"type": "xavier_normal"}],
                ["_span_width_embedding.weight", {"type": "xavier_normal"}],
                ["_context_layer._module.weight_ih.*", {"type": "xavier_normal"}],
                ["_context_layer._module.weight_hh.*", {"type": "orthogonal"}]
            ])
        )
        return model

    def build_dataloader(self,
                         data,
                         batch_size,
                         shuffle,
                         device,
                         logger: logging.Logger,
                         **kwargs) -> DataLoader:
        dataset = CONLL12CorefDataset(data, [FieldLength('text')])
        if isinstance(self.config.embed, Embedding):
            transform = self.config.embed.transform(vocabs=self.vocabs)
            if transform:
                dataset.append_transform(transform)
        dataset.append_transform(self.vocabs)
        if isinstance(data, str):
            dataset.purge_cache()  # Enable cache
        if self.vocabs.mutable:
            self.build_vocabs(dataset)
        return PadSequenceDataLoader(batch_size=batch_size,
                                     shuffle=shuffle,
                                     device=device,
                                     dataset=dataset,
                                     pad={'spans': 0, 'span_labels': -1})

    def predict(self, data: Union[str, List[str]], batch_size: int = None, **kwargs):
        pass

    # noinspection PyMethodOverriding
    def fit(self,
            trn_data,
            dev_data,
            save_dir,
            batch_size,
            embed: Embedding,
            mention_feedforward: FeedForward,
            antecedent_feedforward: FeedForward,
            feature_size: int,
            max_span_width: int,
            spans_per_word: float,
            max_antecedents: int,
            lr=1e-3,
            transformer_lr=1e-5,
            adam_epsilon=1e-6,
            weight_decay=0.01,
            warmup_steps=0.1,
            epochs=150,
            grad_norm=None,
            coarse_to_fine: bool = False,
            inference_order: int = 1,
            lexical_dropout: float = 0.2,
            context_layer: LSTMContextualEncoder = None,
            devices=None,
            logger=None,
            seed=None,
            **kwargs
            ):
        return super().fit(**merge_locals_kwargs(locals(), kwargs))

    def update_metric(self, metric, logits: torch.Tensor, target, output=None):
        metric(logits, target)
        if output:
            label_ids = logits.argmax(-1)
            return label_ids

    def compute_loss(self, criterion, logits, target, batch):
        loss = criterion(logits, target)
        return loss

    def feed_batch(self, batch) -> Dict[str, Any]:
        output_dict = self.model(batch)
        return output_dict

    def build_vocabs(self, dataset, **kwargs):
        if self.vocabs:
            for each in dataset:
                pass
            self.vocabs.lock()
            self.vocabs.summary()

    def reset_metrics(self):
        for each in self.build_metric():
            each.reset()

    def report_metrics(self, loss):
        mention_recall, conll_coref_scores = self.build_metric()
        return f'loss:{loss:.4f} mention_recall:{mention_recall.get_metric():.2%} f1:{conll_coref_scores.get_metric()[-1]:.2%}'



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-11-10 11:47
from hanlp.common.dataset import TableDataset

STANFORD_SENTIMENT_TREEBANK_2_TRAIN = 'http://file.hankcs.com/corpus/SST2.zip#train.tsv'
STANFORD_SENTIMENT_TREEBANK_2_DEV = 'http://file.hankcs.com/corpus/SST2.zip#dev.tsv'
STANFORD_SENTIMENT_TREEBANK_2_TEST = 'http://file.hankcs.com/corpus/SST2.zip#test.tsv'

MICROSOFT_RESEARCH_PARAPHRASE_CORPUS_TRAIN = 'http://file.hankcs.com/corpus/mrpc.zip#train.tsv'
MICROSOFT_RESEARCH_PARAPHRASE_CORPUS_DEV = 'http://file.hankcs.com/corpus/mrpc.zip#dev.tsv'
MICROSOFT_RESEARCH_PARAPHRASE_CORPUS_TEST = 'http://file.hankcs.com/corpus/mrpc.zip#test.tsv'


class SST2Dataset(TableDataset):
    pass


def main():
    dataset = SST2Dataset(STANFORD_SENTIMENT_TREEBANK_2_TEST)
    print(dataset)


if __name__ == '__main__':
    main()



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-29 01:57
from hanlp_common.constant import HANLP_URL

CTB5_POS_RNN = HANLP_URL + 'pos/ctb5_pos_rnn_20200113_235925.zip'
'An old school BiLSTM tagging model trained on CTB5.'
CTB5_POS_RNN_FASTTEXT_ZH = HANLP_URL + 'pos/ctb5_pos_rnn_fasttext_20191230_202639.zip'
'An old school BiLSTM tagging model with FastText (:cite:`bojanowski2017enriching`) embeddings trained on CTB5.'
CTB9_POS_ALBERT_BASE = HANLP_URL + 'pos/ctb9_albert_base_zh_epoch_20_20201011_090522.zip'
'ALBERT model (:cite:`Lan2020ALBERT:`) trained on CTB9.'

PTB_POS_RNN_FASTTEXT_EN = HANLP_URL + 'pos/ptb_pos_rnn_fasttext_20200103_145337.zip'
'An old school BiLSTM tagging model with FastText (:cite:`bojanowski2017enriching`) embeddings trained on PTB.'

ALL = {}


# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2020-07-22 23:40
import logging
from typing import Union, List, Callable, Any, Dict

import torch
from torch.utils.data import DataLoader

from hanlp.common.dataset import PadSequenceDataLoader
from hanlp.common.structure import History
from hanlp.common.torch_component import TorchComponent
from hanlp.common.transform import FieldLength
from hanlp.common.vocab import Vocab
from hanlp.components.ner.biaffine_ner.biaffine_ner import BiaffineNamedEntityRecognizer
from hanlp.components.parsers.hpsg import trees, bracket_eval, dep_eval
from hanlp.components.parsers.hpsg.bracket_eval import FScore
from hanlp.components.parsers.hpsg.dep_eval import SimpleAttachmentScore
from hanlp.components.parsers.hpsg.hpsg_dataset import HeadDrivenPhraseStructureDataset
from hanlp.components.parsers.hpsg.hpsg_parser_model import ChartParser
from hanlp.datasets.parsing.conll_dataset import append_bos_eos
from hanlp.layers.embeddings.embedding import Embedding
from hanlp.metrics.parsing.attachmentscore import AttachmentScore
from hanlp.utils.time_util import CountdownTimer
from hanlp_common.util import merge_locals_kwargs


class HeadDrivenPhraseStructureParser(TorchComponent):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.model: ChartParser = None

    # noinspection PyCallByClass
    def build_optimizer(self,
                        trn,
                        epochs,
                        lr,
                        adam_epsilon,
                        weight_decay,
                        warmup_steps,
                        transformer_lr,
                        **kwargs):
        return BiaffineNamedEntityRecognizer.build_optimizer(self,
                                                             trn,
                                                             epochs,
                                                             lr,
                                                             adam_epsilon,
                                                             weight_decay,
                                                             warmup_steps,
                                                             transformer_lr)

    def build_criterion(self, **kwargs):
        pass

    def build_metric(self, **kwargs):
        return AttachmentScore()

    def build_model(self, training=True, **kwargs) -> torch.nn.Module:
        model = ChartParser(self.config.embed.module(vocabs=self.vocabs),
                            self.vocabs.pos, self.vocabs.label, self.vocabs.rel, self.config)
        return model

    # noinspection PyMethodOverriding
    def build_dataloader(self, data, batch_size, shuffle, device, logger: logging.Logger, sampler_builder,
                         gradient_accumulation,
                         **kwargs) -> DataLoader:
        # shuffle = False  # We need to find the smallest grad_acc
        dataset = HeadDrivenPhraseStructureDataset(data, transform=[append_bos_eos])
        if self.config.get('transform', None):
            dataset.append_transform(self.config.transform)
        dataset.append_transform(self.vocabs)
        if isinstance(self.config.embed, Embedding):
            transform = self.config.embed.transform(vocabs=self.vocabs)
            if transform:
                dataset.append_transform(transform)
        dataset.append_transform(self.vocabs)
        field_length = FieldLength('token')
        dataset.append_transform(field_length)
        if isinstance(data, str):
            dataset.purge_cache()  # Enable cache
        if self.vocabs.mutable:
            self.build_vocabs(dataset, logger)
        if 'token' in self.vocabs:
            lens = [x[field_length.dst] for x in dataset]
        else:
            lens = [len(x['token_input_ids']) for x in dataset]
        if sampler_builder:
            sampler = sampler_builder.build(lens, shuffle, gradient_accumulation)
        else:
            sampler = None
        return PadSequenceDataLoader(batch_sampler=sampler,
                                     batch_size=batch_size,
                                     device=device,
                                     dataset=dataset)

    def predict(self, data: Union[str, List[str]], batch_size: int = None, **kwargs):
        pass

    def build_vocabs(self, dataset, logger, **kwargs):
        self.vocabs.rel = Vocab(pad_token=None, unk_token=None)
        self.vocabs.pos = Vocab(pad_token=None, unk_token=None)
        self.vocabs.label = label_vocab = Vocab(pad_token='', unk_token=None)
        label_vocab.add(trees.Sub_Head)
        for each in dataset:
            tree = each['hpsg']
            nodes = [tree]
            while nodes:
                node = nodes.pop()
                if isinstance(node, trees.InternalParseNode):
                    label_vocab.add('\t'.join(node.label))
                    nodes.extend(reversed(node.children))
        self.vocabs['rel'].set_unk_as_safe_unk()
        label_vocab.set_unk_as_safe_unk()
        self.vocabs.lock()
        self.vocabs.summary(logger)

    def fit(self, trn_data, dev_data, save_dir,
            embed: Embedding,
            batch_size=100,
            epochs=100,
            sampler='sorting',
            n_buckets=32,
            batch_max_tokens=None,
            sampler_builder=None,
            attention_dropout=0.2,
            bert_do_lower_case=True,
            bert_model='bert-large-uncased',
            bert_transliterate='',
            char_lstm_input_dropout=0.2,
            clip_grad_norm=0.0,
            const_lada=0.5,
            d_biaffine=1024,
            d_char_emb=64,
            d_ff=2048,
            d_kv=64,
            d_label_hidden=250,
            d_model=1024,
            dataset='ptb',
            elmo_dropout=0.5,
            embedding_dropout=0.2,
            embedding_path='data/glove.gz',
            embedding_type='random',
            lal_combine_as_self=False,
            lal_d_kv=128,
            lal_d_proj=128,
            lal_partitioned=True,
            lal_pwff=True,
            lal_q_as_matrix=False,
            lal_resdrop=False,
            max_len_dev=0,
            max_len_train=0,
            morpho_emb_dropout=0.2,
            num_heads=8,
            num_layers=3,
            pad_left=False,
            partitioned=True,
            relu_dropout=0.2,
            residual_dropout=0.2,
            sentence_max_len=300,
            step_decay=True,
            step_decay_factor=0.5,
            step_decay_patience=5,
            tag_emb_dropout=0.2,
            timing_dropout=0.0,
            dont_use_encoder=False,
            use_cat=False,
            use_chars_lstm=False,
            use_elmo=False,
            use_lal=True,
            use_tags=True,
            use_words=False,
            word_emb_dropout=0.4,
            lr=1e-3,
            transformer_lr=5e-5,
            adam_epsilon=1e-6,
            weight_decay=0.01,
            warmup_steps=0.1,
            grad_norm=5.0,
            gradient_accumulation=1,
            devices=None, logger=None, seed=None, **kwargs):
        return super().fit(**merge_locals_kwargs(locals(), kwargs))

    # noinspection PyMethodOverriding
    def fit_dataloader(self,
                       trn: DataLoader,
                       criterion,
                       optimizer,
                       metric,
                       logger: logging.Logger,
                       history: History,
                       linear_scheduler=None,
                       gradient_accumulation=1,
                       **kwargs):
        self.model.train()
        timer = CountdownTimer(history.num_training_steps(len(trn), gradient_accumulation))
        total_loss = 0
        self.reset_metrics(metric)
        for idx, batch in enumerate(trn):
            output_dict = self.feed_batch(batch)
            self.update_metrics(batch, output_dict, metric)
            loss = output_dict['loss']
            if gradient_accumulation > 1:
                loss /= gradient_accumulation
            loss.backward()
            total_loss += loss.item()
            if history.step(gradient_accumulation):
                self._step(optimizer, linear_scheduler)
                timer.log(self.report_metrics(total_loss / (timer.current + 1), metric), ratio_percentage=None,
                          logger=logger)
            del loss
        return total_loss / timer.total

    def _step(self, optimizer, linear_scheduler):
        if self.config.grad_norm:
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.config.grad_norm)
        optimizer.step()
        optimizer.zero_grad()
        if linear_scheduler:
            linear_scheduler.step()

    # noinspection PyMethodOverriding
    def evaluate_dataloader(self,
                            data: DataLoader,
                            criterion: Callable,
                            metric,
                            logger,
                            ratio_width=None,
                            output=False,
                            **kwargs):
        self.model.eval()
        self.reset_metrics(metric)
        timer = CountdownTimer(len(data))
        gold_tree = []
        pred_tree = []
        pred_head = []
        pred_type = []
        gold_type = []
        gold_word = []
        gold_pos = []
        gold_head = []
        for batch in data:
            output_dict = self.feed_batch(batch)
            gold_tree += batch['tree']
            pred_tree += output_dict['predicted_tree']

            pred_head += output_dict['pred_head']
            pred_type += output_dict['pred_type']

            gold_type += batch['DEPREL']
            gold_head += batch['HEAD']
            gold_pos += batch['CPOS']
            gold_word += batch['FORM']
            assert len(pred_head) == len(gold_head)
            self.update_metrics(batch, output_dict, metric)
            timer.log('', ratio_percentage=None, ratio_width=ratio_width)

        tree_score: FScore = bracket_eval.evalb(gold_tree, pred_tree)
        assert len(pred_head) == len(pred_type)
        assert len(pred_type) == len(gold_type)
        lens = [len(x) for x in gold_word]
        stats, stats_nopunc, stats_root, test_total_inst = dep_eval.eval(len(pred_head), gold_word, gold_pos,
                                                                         pred_head,
                                                                         pred_type, gold_head, gold_type,
                                                                         lens, punct_set=None,
                                                                         symbolic_root=False)

        test_ucorrect, test_lcorrect, test_total, test_ucomlpete_match, test_lcomplete_match = stats
        test_ucorrect_nopunc, test_lcorrect_nopunc, test_total_nopunc, test_ucomlpete_match_nopunc, test_lcomplete_match_nopunc = stats_nopunc
        test_root_correct, test_total_root = stats_root
        dep_score = SimpleAttachmentScore(test_ucorrect_nopunc / test_total_nopunc,
                                          test_lcorrect_nopunc / test_total_nopunc)
        timer.log(f'{tree_score} {dep_score}', ratio_percentage=None, ratio_width=ratio_width, logger=logger)
        return tree_score, dep_score

    def reset_metrics(self, metrics):
        pass
        # for m in metrics:
        #     m.reset()

    def report_metrics(self, loss, metrics):
        return f'loss: {loss:.4f}'

    def feed_batch(self, batch) -> Dict[str, Any]:
        predicted_tree, loss_or_score = self.model(batch)
        outputs = {}
        if isinstance(loss_or_score, torch.Tensor):
            loss_or_score /= len(batch['hpsg'])
            loss = loss_or_score
            outputs['loss'] = loss
        else:
            score = loss_or_score
            outputs['score'] = score
        if predicted_tree:
            predicted_tree = [p.convert() for p in predicted_tree]
            pred_head = [[leaf.father for leaf in tree.leaves()] for tree in predicted_tree]
            pred_type = [[leaf.type for leaf in tree.leaves()] for tree in predicted_tree]
            outputs.update({
                'predicted_tree': predicted_tree,
                'pred_head': pred_head,
                'pred_type': pred_type
            }),
        return outputs

    def update_metrics(self, batch: dict, output_dict: dict, metrics):
        pass
        # assert len(output_dict['prediction']) == len(batch['ner'])
        # for pred, gold in zip(output_dict['prediction'], batch['ner']):
        #     metrics(set(pred), set(gold))

    def execute_training_loop(self,
                              trn: DataLoader,
                              dev: DataLoader,
                              epochs,
                              criterion,
                              optimizer,
                              metric,
                              save_dir,
                              logger: logging.Logger,
                              devices,
                              **kwargs):
        best_epoch, best_score = 0, -1
        optimizer, scheduler = optimizer
        timer = CountdownTimer(epochs)
        _len_trn = len(trn) // self.config.gradient_accumulation
        ratio_width = len(f'{_len_trn}/{_len_trn}')
        history = History()
        for epoch in range(1, epochs + 1):
            logger.info(f"[yellow]Epoch {epoch} / {epochs}:[/yellow]")
            self.fit_dataloader(trn, criterion, optimizer, metric, logger, history,
                                linear_scheduler=scheduler if self.use_transformer else None, **kwargs)
            if dev:
                metric = self.evaluate_dataloader(dev, criterion, metric, logger, ratio_width=ratio_width)
            report = f'{timer.elapsed_human}/{timer.total_time_human}'
            dev_score = sum(x.score for x in metric) / len(metric)
            if not self.use_transformer:
                scheduler.step(dev_score)
            if dev_score > best_score:
                self.save_weights(save_dir)
                best_score = dev_score
                report += ' [red]saved[/red]'
            timer.log(report, ratio_percentage=False, newline=True, ratio=False)

    @property
    def use_transformer(self):
        return 'token' not in self.vocabs

    def _get_transformer(self):
        return getattr(self.model.embed, 'transformer', None)



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-08-26 14:45
import inspect
from abc import ABC, abstractmethod
from typing import Any

from hanlp_common.configurable import Configurable


class Component(Configurable, ABC):
    @abstractmethod
    def predict(self, data: Any, **kwargs):
        """Predict on data. This is the base class for all components, including rule based and statistical ones.

        Args:
          data: Any type of data subject to sub-classes
          kwargs: Additional arguments

        Returns: Any predicted annotations.

        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    def __call__(self, data: Any, **kwargs):
        """
        A shortcut for :func:`~hanlp.common.component.predict`.

        Args:
            data:
            **kwargs:

        Returns:

        """
        return self.predict(data, **kwargs)



# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2020-12-07 21:45
import os
import shutil

from hanlp.components.parsers.ud.udify_util import get_ud_treebank_files
from hanlp.utils.io_util import get_resource
from hanlp.utils.log_util import flash


def concat_treebanks(home, version):
    ud_home = get_resource(home)
    treebanks = get_ud_treebank_files(ud_home)
    output_dir = os.path.abspath(os.path.join(ud_home, os.path.pardir, os.path.pardir, f'ud-multilingual-v{version}'))
    if os.path.isdir(output_dir):
        return output_dir
    os.makedirs(output_dir)
    train, dev, test = list(zip(*[treebanks[k] for k in treebanks]))

    for treebank, name in zip([train, dev, test], ["train.conllu", "dev.conllu", "test.conllu"]):
        flash(f'Concatenating {len(train)} treebanks into {name} [blink][yellow]...[/yellow][/blink]')
        with open(os.path.join(output_dir, name), 'w') as write:
            for t in treebank:
                if not t:
                    continue
                with open(t, 'r') as read:
                    shutil.copyfileobj(read, write)
        flash('')
    return output_dir
